"""Autogenerated version information file.

Manual changes to this file will be overwritten.

The output of git describe was: ``2.0.3``

The derived version is: ``2.0.3``
"""
__version__ = '2.0.3'
DEV_CLASSIFIER = 'Development Status :: 5 - Production/Stable'
